import os
import h5py
from sklearn.model_selection import train_test_split
from tensorflow.keras.utils import to_categorical
import pickle
import multiprocessing as mp
import numpy as np
import pandas as pd
import itertools
import copy
from sklearn.utils import shuffle
from scipy.interpolate import interp1d

from astrorapid import helpers
from astrorapid.get_training_data import get_data

# fix random seed for reproducibility
np.random.seed(42)


class PrepareArrays(object):
    def __init__(self, passbands=('g', 'r'), contextual_info=('redshift',), nobs=50, mintime=-70, maxtime=80,
                 timestep=3.0):
        self.passbands = passbands
        self.contextual_info = contextual_info
        self.npassbands = len(passbands)
        self.nfeatures = self.npassbands + len(self.contextual_info)
        self.nobs = nobs
        self.timestep = timestep
        self.mintime = mintime
        self.maxtime = maxtime

    def make_cuts(self, data, i, deleterows, b, redshift=None, class_num=None, bcut=True, zcut=0.5, ignore_classes=(),
                  pre_trigger=True):
        deleted = False
        try:
            time = data[data['passband']=='r']['time'].data
        except KeyError:
            print("No r band data. passbands")
            deleterows.append(i)
            deleted = True
            return deleterows, deleted

        if len(data) < 4:
            print("Less than 4 epochs. nobs = {}".format(len(data)))
            deleterows.append(i)
            deleted = True
        elif pre_trigger and len(time[time < 0]) < 3:
            print("Less than 3 points in the r band pre trigger", len(time[time < 0]))
            deleterows.append(i)
            deleted = True
        elif bcut and abs(b) < 15:
            print("In galactic plane. b = {}".format(b))
            deleterows.append(i)
            deleted = True
        elif zcut is not None and redshift is not None and (redshift > self.zcut or redshift == 0):
            print("Redshift cut. z = {}".format(redshift))
            deleterows.append(i)
            deleted = True
        elif class_num in ignore_classes:
            print("Not including class:", class_num)
            deleterows.append(i)
            deleted = True

        return deleterows, deleted

    def get_min_max_time(self, data):
        # Get min and max times for tinterp
        mintimes = []
        maxtimes = []
        for j, pb in enumerate(self.passbands):
            pbmask = data['passband']==pb
            time = data[pbmask]['time'].data
            try:
                mintimes.append(time.min())
                maxtimes.append(time.max())
            except ValueError:
                print("No data for passband: ", pb)
        mintime = min(mintimes)
        maxtime = max(maxtimes) + self.timestep

        return mintime, maxtime

    def get_t_interp(self, data):
        mintime, maxtime = self.get_min_max_time(data)

        tinterp = np.arange(mintime, maxtime, step=self.timestep)
        len_t = len(tinterp)
        if len_t > self.nobs:
            tinterp = tinterp[(tinterp >= self.mintime)]
            len_t = len(tinterp)
            if len_t > self.nobs:
                tinterp = tinterp[:-(len_t - self.nobs)]
                len_t = len(tinterp)
        return tinterp, len_t

    def update_X(self, X, i, data, tinterp, len_t, objid, contextual_info, meta_data):
        for j, pb in enumerate(self.passbands):
            # Drop infinite or nan values in any row
            data.remove_rows(np.where(~np.isfinite(data['time']))[0])
            data.remove_rows(np.where(~np.isfinite(data['flux']))[0])
            data.remove_rows(np.where(~np.isfinite(data['fluxErr']))[0])

            # Get data
            pbmask = data['passband']==pb
            time = data[pbmask]['time'].data
            flux = data[pbmask]['flux'].data
            fluxerr = data[pbmask]['fluxErr'].data
            photflag = data[pbmask]['photflag'].data

            # Mask out times outside of mintime and maxtime
            timemask = (time > self.mintime) & (time < self.maxtime)
            time = time[timemask]
            flux = flux[timemask]
            fluxerr = fluxerr[timemask]
            photflag = photflag[timemask]

            n = len(flux)  # Get vector length (could be less than nobs)

            if n > 1:
                # if flux[-1] > flux[-2]:  # If last values are increasing, then set fill_values to zero
                #     f = interp1d(time, flux, kind='linear', bounds_error=False, fill_value=0.)
                # else:
                #     f = interp1d(time, flux, kind='linear', bounds_error=False,
                #                  fill_value='extrapolate')  # extrapolate until all passbands finished.
                f = interp1d(time, flux, kind='linear', bounds_error=False, fill_value=0.)

                fluxinterp = f(tinterp)
                fluxinterp = np.nan_to_num(fluxinterp)
                fluxinterp = fluxinterp.clip(min=0)
                fluxerrinterp = np.zeros(len_t)

                for interp_idx, fluxinterp_val in enumerate(fluxinterp):
                    if fluxinterp_val == 0.:
                        fluxerrinterp[interp_idx] = 0
                    else:
                        nearest_idx = helpers.find_nearest(time, tinterp[interp_idx])
                        fluxerrinterp[interp_idx] = fluxerr[nearest_idx]

                X[i][j][0:len_t] = fluxinterp
                # X[i][j * 2 + 1][0:len_t] = fluxerrinterp

        # Add contextual information
        for jj, c_info in enumerate(contextual_info, 1):
            X[i][j + jj][0:len_t] = meta_data[c_info] * np.ones(len_t)

        return X


class PrepareInputArrays(PrepareArrays):
    def __init__(self, passbands=('g', 'r'), contextual_info=('redshift',), bcut=True, zcut=None,
                 nobs=50, mintime=-70, maxtime=80, timestep=3.0):
        PrepareArrays.__init__(self, passbands, contextual_info, nobs, mintime, maxtime, timestep)
        self.bcut = bcut
        self.zcut = zcut

    def prepare_input_arrays(self, lightcurves):
        nobjects = len(lightcurves)

        X = np.zeros(shape=(nobjects, self.nfeatures, self.nobs))
        timesX = np.zeros(shape=(nobjects, self.nobs))
        objids_list = []
        orig_lc = []
        deleterows = []
        trigger_mjds = []

        for i, (objid, data) in enumerate(lightcurves.items()):
            print("Preparing light curve {} of {}".format(i, nobjects))

            redshift = data.meta['redshift']
            b = data.meta['b']
            trigger_mjd = data.meta['trigger_mjd']

            # Make cuts
            deleterows, deleted = self.make_cuts(data, i, deleterows, b, redshift, class_num=None, bcut=self.bcut,
                                                 zcut=self.zcut, pre_trigger=False)
            if deleted:
                continue

            tinterp, len_t = self.get_t_interp(data)
            timesX[i][0:len_t] = tinterp
            orig_lc.append(data)
            objids_list.append(objid)
            trigger_mjds.append(trigger_mjd)
            X = self.update_X(X, i, data, tinterp, len_t, objid, self.contextual_info, data.meta)

        deleterows = np.array(deleterows)
        X = np.delete(X, deleterows, axis=0)
        timesX = np.delete(timesX, deleterows, axis=0)

        # Correct shape for keras is (N_objects, N_timesteps, N_passbands) (where N_timesteps is lookback time)
        X = X.swapaxes(2, 1)

        return X, orig_lc, timesX, objids_list, trigger_mjds


class PrepareTrainingSetArrays(PrepareArrays):
    def __init__(self, passbands=('g', 'r'), contextual_info=('redshift',), nobs=50, mintime=-70, maxtime=80,
                 timestep=3.0, reread=False, bcut=True, zcut=None, ignore_classes=(), class_name_map=None,
                 nchunks=10000,  training_set_dir='data/training_set_files', data_dir='data/ZTF_20190512/',
                 save_dir='data/saved_light_curves/', get_data_func=None):
        PrepareArrays.__init__(self, passbands, contextual_info, nobs, mintime, maxtime, timestep)
        self.reread = reread
        self.bcut = bcut
        self.zcut = zcut
        self.ignore_classes = ignore_classes
        self.nchunks = nchunks
        self.training_set_dir = training_set_dir
        self.data_dir = data_dir
        self.save_dir = save_dir
        self.light_curves = {}
        self.get_data_func = get_data_func
        if 'redshift' in contextual_info:
            self.known_redshift = True
        else:
            self.known_redshift = False
        if class_name_map is None:
            self.class_name_map = helpers.get_sntypes()
        else:
            self.class_name_map = class_name_map

        if not os.path.exists(self.training_set_dir):
            os.makedirs(self.training_set_dir)

    def get_light_curves(self, class_nums=(1,), nprocesses=1):
        light_curves = {}

        for class_num in class_nums:
            lcs = get_data(self.get_data_func, class_num, self.data_dir, self.save_dir, self.passbands,
                           self.known_redshift, nprocesses, self.reread)
            light_curves.update(lcs)

        return light_curves

    def prepare_training_set_arrays(self, otherchange='', class_nums=(1,), nprocesses=1, train_size=0.6):
        savepath = os.path.join(self.training_set_dir,
                                "X_{}ci{}_z{}_b{}_ig{}.npy".format(otherchange, self.contextual_info, self.zcut,
                                                                         self.bcut, self.ignore_classes))
        print(savepath)
        if self.reread is True or not os.path.isfile(savepath):
            self.light_curves = self.get_light_curves(class_nums, nprocesses)

            objids = list(set(self.light_curves.keys()))
            nobjects = len(objids)

            # Store data labels (y) and 'r' band data (X). Use memory mapping because input file is very large.
            labels = np.empty(shape=nobjects, dtype=object)
            y = np.zeros(shape=(nobjects, self.nobs), dtype=object)
            X = np.memmap(os.path.join(self.training_set_dir, 'X_lc_data.dat'), dtype=np.float32, mode='w+',
                          shape=(nobjects, self.nfeatures, self.nobs))  # 4+len(self.contextual_info), 100))
            X[:] = np.zeros(shape=(nobjects, self.nfeatures, self.nobs))
            timesX = np.zeros(shape=(nobjects, self.nobs))
            objids_list = []
            orig_lc = []

            # Chunk before multiprocessing
            multi_objids = np.array_split(objids, self.nchunks)

            # Store light curves into X (fluxes) and y (labels)
            if nprocesses == 1:
                outputs = []
                for arg in multi_objids:
                    outputs.append(self.multi_read_obj(arg))
            else:
                pool = mp.Pool(nprocesses)
                results = pool.map_async(self.multi_read_obj, multi_objids) ##
                pool.close()
                pool.join()
                outputs = results.get()

            sum_deleterows = 0
            startidx = 0
            num_outputs = len(outputs)
            print('combining results...')
            for i, output in enumerate(outputs):
                labels_part, y_part, X_part, timesX_part, objids_list_part, orig_lc_part, num_deleterows_part, num_objects_part = output
                endidx = startidx + num_objects_part
                labels[startidx:endidx] = labels_part
                y[startidx:endidx] = y_part
                X[startidx:endidx] = X_part
                timesX[startidx:endidx] = timesX_part
                objids_list.extend(objids_list_part)
                orig_lc.extend(orig_lc_part)
                startidx += num_objects_part
                sum_deleterows += num_deleterows_part

            deleterows = np.array(np.arange(nobjects - sum_deleterows, nobjects))
            X = np.delete(X, deleterows, axis=0)
            y = np.delete(y, deleterows, axis=0)
            labels = np.delete(labels, deleterows, axis=0)
            timesX = np.delete(timesX, deleterows, axis=0)

            np.save(os.path.join(self.training_set_dir,
                                 "X_{}ci{}_z{}_b{}_ig{}.npy".format(otherchange, self.contextual_info, self.zcut,
                                                                    self.bcut, self.ignore_classes)), X)
            np.save(os.path.join(self.training_set_dir,
                                 "y_{}ci{}_z{}_b{}_ig{}.npy".format(otherchange, self.contextual_info, self.zcut,
                                                                    self.bcut, self.ignore_classes)), y,
                    allow_pickle=True)
            np.save(os.path.join(self.training_set_dir,
                                 "labels_{}ci{}_z{}_b{}_ig{}.npy".format(otherchange, self.contextual_info, self.zcut,
                                                                         self.bcut, self.ignore_classes)), labels,
                    allow_pickle=True)
            np.save(os.path.join(self.training_set_dir,
                                 "tinterp_{}ci{}_z{}_b{}_ig{}.npy".format(otherchange, self.contextual_info, self.zcut,
                                                                          self.bcut, self.ignore_classes)), timesX)
            np.save(os.path.join(self.training_set_dir,
                                 "objids_{}ci{}_z{}_b{}_ig{}.npy".format(otherchange, self.contextual_info, self.zcut,
                                                                         self.bcut, self.ignore_classes)), objids_list,
                    allow_pickle=True)
            with open(os.path.join(self.training_set_dir,
                                   "origlc_{}ci{}_z{}_b{}_ig{}.npy".format(otherchange, self.contextual_info, self.zcut,
                                                                           self.bcut, self.ignore_classes)),
                      'wb') as f:
                pickle.dump(orig_lc, f)

        else:
            X = np.load(os.path.join(self.training_set_dir,
                                     "X_{}ci{}_z{}_b{}_ig{}.npy".format(otherchange, self.contextual_info, self.zcut,
                                                                        self.bcut, self.ignore_classes)), mmap_mode='r')
            y = np.load(os.path.join(self.training_set_dir,
                                     "y_{}ci{}_z{}_b{}_ig{}.npy".format(otherchange, self.contextual_info, self.zcut,
                                                                        self.bcut, self.ignore_classes)),
                        allow_pickle=True)
            labels = np.load(os.path.join(self.training_set_dir,
                                          "labels_{}ci{}_z{}_b{}_ig{}.npy".format(otherchange, self.contextual_info,
                                                                                  self.zcut, self.bcut,
                                                                                  self.ignore_classes)),
                             allow_pickle=True)
            timesX = np.load(os.path.join(self.training_set_dir,
                                          "tinterp_{}ci{}_z{}_b{}_ig{}.npy".format(otherchange, self.contextual_info,
                                                                                   self.zcut, self.bcut,
                                                                                   self.ignore_classes)))
            objids_list = np.load(os.path.join(self.training_set_dir,
                                               "objids_{}ci{}_z{}_b{}_ig{}.npy".format(otherchange,
                                                                                       self.contextual_info, self.zcut,
                                                                                       self.bcut, self.ignore_classes)),
                                  allow_pickle=True)
            with open(os.path.join(self.training_set_dir,
                                   "origlc_{}ci{}_z{}_b{}_ig{}.npy".format(otherchange, self.contextual_info, self.zcut,
                                                                           self.bcut, self.ignore_classes)), 'rb') as f:
                orig_lc = pickle.load(f)

        classes = sorted(list(set(labels)))

        # Count nobjects per class
        for c in classes:
            nobs = len(X[labels == c])
            print(c, nobs)

        # Use class numbers 1,2,3... instead of 1, 3, 13 etc.
        y_indexes = np.copy(y)
        for i, c in enumerate(classes):
            y_indexes[y == c] = i + 1
        y = y_indexes

        y = to_categorical(y)

        # Correct shape for keras is (N_objects, N_timesteps, N_passbands) (where N_timesteps is lookback time)
        X = X.swapaxes(2, 1)

        # #NORMALISE
        # X = X.copy()
        # for i in range(len(X)):
        #     for pbidx in range(2):
        #         minX = X[i, :, pbidx].min(axis=0)
        #         maxX = X[i, :, pbidx].max(axis=0)
        #         X[i, :, pbidx] = (X[i, :, pbidx] - minX) / (maxX - minX)
        #         # if (maxX - minX) != 0:
        #         #     mask.append(i)
        #         #     break
        # finitemask = ~np.any(np.any(~np.isfinite(X), axis=1), axis=1)
        # X = X[finitemask]
        # y = y[finitemask]
        # timesX = timesX[finitemask]
        # objids_list = objids_list[finitemask]
        # orig_lc = list(itertools.compress(orig_lc, finitemask))
        # labels = labels[finitemask]

        print("Shuffling")
        X, y, labels, timesX, orig_lc, objids_list = shuffle(X, y, labels, timesX, orig_lc, objids_list)
        print("Done shuffling")

        X_train, X_test, y_train, y_test, labels_train, labels_test, timesX_train, timesX_test, orig_lc_train, \
        orig_lc_test, objids_train, objids_test = train_test_split(
            X, y, labels, timesX, orig_lc, objids_list, train_size=train_size, shuffle=False, random_state=42)

        def augment_crop_lightcurves(X_local, y_local, labels_local, timesX_local, orig_lc_local, objids_local):
            X_local = copy.copy(X_local)
            y_local = copy.copy(y_local)
            labels_local = copy.copy(labels_local)
            timesX_local = copy.copy(timesX_local)
            orig_lc_local = copy.copy(orig_lc_local)
            objids_local = copy.copy(objids_local)

            newX = np.zeros(X_local.shape)
            newy = np.zeros(y_local.shape)
            lenX = len(X_local)
            for i in range(lenX):
                if i % 1000 == 0:
                    print(f"new {i} of {lenX}")
                mask = timesX_local[i] >= 0
                nmask = sum(mask)
                newX[i][:nmask] = X_local[i][mask]
                newy[i][:nmask] = y_local[i][mask]

            print("Concatenating")
            X_local = np.concatenate((X_local, newX))
            y_local = np.concatenate((y_local, newy))
            labels_local = np.concatenate((labels_local, labels_local))
            timesX_local = np.concatenate((timesX_local, timesX_local))
            orig_lc_local = orig_lc_local * 2
            objids_local = np.concatenate((objids_local, objids_local))

            return X_local, y_local, labels_local, timesX_local, orig_lc_local, objids_local

        X_train, y_train, labels_train, timesX_train, orig_lc_train, objids_train = augment_crop_lightcurves(X_train, y_train, labels_train, timesX_train, orig_lc_train, objids_train)
        X_test, y_test, labels_test, timesX_test, orig_lc_test, objids_test = augment_crop_lightcurves(X_test, y_test, labels_test, timesX_test, orig_lc_test, objids_test)

        X_train, y_train, labels_train, timesX_train, orig_lc_train, objids_train = shuffle(X_train, y_train, labels_train, timesX_train, orig_lc_train, objids_train)

        counts = np.unique(labels_train, return_counts=True)[-1]
        class_weights = max(counts) / counts
        class_weights = dict(zip(range(len(counts)), class_weights))
        print("Class weights:", class_weights)

        # Sample weights
        l_train_indexes = np.copy(labels_train)
        for i, c in enumerate(classes):
            l_train_indexes[l_train_indexes == c] = i
        sample_weights = np.zeros(len(l_train_indexes))
        for key, val in class_weights.items():
            sample_weights[l_train_indexes == key] = val

        return X_train, X_test, y_train, y_test, labels_train, labels_test, classes, class_weights, \
               sample_weights, timesX_train, timesX_test, orig_lc_train, orig_lc_test, objids_train, objids_test

    def multi_read_obj(self, objids):
        nobjects = len(objids)

        labels = np.empty(shape=nobjects, dtype=object)
        y = np.zeros(shape=(nobjects, self.nobs), dtype=object)
        X = np.zeros(shape=(nobjects, self.nfeatures, self.nobs))
        timesX = np.zeros(shape=(nobjects, self.nobs))
        objids_list = []
        orig_lc = []
        deleterows = []

        for i, objid in enumerate(objids):
            print("Preparing {} light curve {} of {}".format(objid, i, nobjects))

            # Get data for each object
            data = self.light_curves[objid]

            redshift = data.meta['redshift']
            b = data.meta['b']
            t0 = data.meta['t0']
            class_num = data.meta['class_num']

            # Make cuts
            deleterows, deleted = self.make_cuts(data, i, deleterows, b, redshift, class_num=class_num, bcut=self.bcut,
                                                 zcut=self.zcut, ignore_classes=self.ignore_classes, pre_trigger=False)
            if deleted:
                continue

            tinterp, len_t = self.get_t_interp(data)
            timesX[i][0:len_t] = tinterp
            orig_lc.append(data)
            objids_list.append(objid)
            X = self.update_X(X, i, data, tinterp, len_t, objid, self.contextual_info, data.meta)

            class_name = self.class_name_map[class_num]
            activeindexes = (tinterp > t0)
            labels[i] = class_name
            y[i][0:len_t][activeindexes] = class_name

        deleterows = np.array(deleterows)
        X = np.delete(X, deleterows, axis=0)
        y = np.delete(y, deleterows, axis=0)
        labels = np.delete(labels, deleterows, axis=0)
        timesX = np.delete(timesX, deleterows, axis=0)
        count_deleterows = len(deleterows)
        num_objects = X.shape[0]

        return labels, y, X, timesX, objids_list, orig_lc, count_deleterows, num_objects




import os
from astrorapid.prepare_arrays import PrepareTrainingSetArrays
from astrorapid.plot_metrics import plot_metrics
from astrorapid.neural_network_model import train_model
import astrorapid.get_custom_data


def create_custom_classifier(get_data_func, data_dir, class_nums=(1,2,), class_name_map=None, reread_data=False, train_size=0.6,
                             contextual_info=('redshift',), passbands=('g', 'r'), nobs=50, mintime=-70, maxtime=80,
                             timestep=3.0, retrain_network=False, train_epochs=50, zcut=0.5, bcut=True,
                             ignore_classes=(), nprocesses=1, nchunks=1000, otherchange='',
                             training_set_dir='data/training_set_files', save_dir='data/saved_light_curves',
                             fig_dir='Figures', plot=True):

    """
    Create a classifier with your own data and own training parameters.

    Parameters
    ----------
    get_data_func : func
        Function that reads your data and returns a light_curves object.
        It takes at least the arguments: class_num, data_dir, save_dir, passbands, known_redshift.
        See get_custom_data.py for an example template.
        E.g. get_data_func = get_data_custom(class_num, data_dir, save_dir, passbands)
    data_dir : str
        Directory where data is stored. E.g. data_dir='data/ZTF_20190512/'
    class_nums : tuple
        Class numbers (or names) to train on. E.g. class_nums=(1, 5, 6, 12, 41, 43, 51)
    class_name_map : dict or None
        This maps the class_nums onto class names.
        E.g. class_name_map = {1: 'SNIa-norm', 5: 'SNIbc', 6: 'SNII', 12: 'SNII', 41: 'SNIa-91bg', 43: 'SNIa-x', 51: 'Kilonova'}.
        You may use the same value for a different key if you want the classifier to join these two class_nums under the same label.
        If this is None, it will use the default mapping listed in get_sntypes in helpers.py.
    reread_data : bool
        If this is True, then it will reread your data and resave the processed files, otherwise
        it will check if the data has already been read, processed and saved.
    train_size : float
        Fraction of data to use for training, the remainder will be used for testing/validation.
        Usually choose a value between 0.5 - 1.0 depending on how much data you have.
    contextual_info : tuple of strings
        What contextual information to use while training. You can write any string in this tuple provided that it
        is stored as a keyword in the metadata of the light_curve Table object returned by the `get_data_func`.
    passbands : tuple of str
        Passbands to use.
        E.g. passbands=('g', 'r')
    nobs : int
        Number of points to use in interpolation of light curve between mintime and maxtime.
    mintime : int
        Days from trigger (minimum) to extract from light curve.
    maxtime : int
        Days from trigger (maximum) to extract from light curve.
    timestep : float
        Time-step between interpolated points in light curve.
    retrain_network : bool
        Whether to retrain the neural network or to use the saved network model.
    train_epochs : int
        Number of epochs to train the neural network.
        This is the number of times the nerual network sees each datum in the training set.
        The higher this is the better the classifier will find a local minimum, however, too high, and it might
        overfit and not generalise well to new data
    zcut : float
        Do not train on objects with redshifts higher than zcut.
    bcut : bool
        If True, do not train on objects within 15 degrees of the galactic plane
    ignore_classes : tuple
        Will not train or test on classes listed in this tuple.
    nprocesses : int or None
        Number of computer processes to use while processing the data.
        If None, it will use all the available processors from os.cpu_count().
    nchunks : int
        Number of chunks to split the data set into before doing multiprocessing.
        This should be a small fraction of the number of total objects.
    other_change : str
        A change in this text will signify that a change has been made to one of these training parameters
        and that the data should be resaved and the model retrained should resave the data and retrained.
    training_set_dir : str
        Name of directory to save the data that will be read by the neural network.
    save_dir : str
        Name of directory to save the processes data files.
    fig_dir : str
        Name of directory to save the Classifier metric figures such as confusion matrices.
    plot : bool
        Whether to plot classier metrics such as confusion matrices after training.
    """

    for dirname in [training_set_dir, data_dir, save_dir]:
        if not os.path.exists(dirname):
            os.makedirs(dirname)

    for dirname in [fig_dir, fig_dir + '/cf_since_trigger', fig_dir + '/cf_since_t0', fig_dir + '/roc_since_trigger',
                fig_dir + '/lc_pred', fig_dir + '/pr_since_trigger', fig_dir + '/truth_table_since_trigger']:
        if not os.path.exists(dirname):
            os.makedirs(dirname)

    # Prepare the training set, read the data files, and save processed files
    preparearrays = PrepareTrainingSetArrays(passbands, contextual_info, nobs, mintime, maxtime, timestep,
                                             reread_data, bcut, zcut, ignore_classes,
                                             class_name_map=class_name_map, nchunks=nchunks,
                                             training_set_dir=training_set_dir, data_dir=data_dir, save_dir=save_dir,
                                             get_data_func=get_data_func)
    X_train, X_test, y_train, y_test, labels_train, labels_test, class_names, class_weights, sample_weights, \
    timesX_train, timesX_test, orig_lc_train, orig_lc_test, objids_train, objids_test = \
        preparearrays.prepare_training_set_arrays(otherchange, class_nums, nprocesses, train_size)

    # Train the neural network model on saved files
    model = train_model(X_train, X_test, y_train, y_test, sample_weights=sample_weights, fig_dir=fig_dir,
                        retrain=retrain_network, epochs=train_epochs, plot_loss=plot)

    # Plot classification metrics such as confusion matrices
    if plot:
        plot_metrics(class_names, model, X_test, y_test, fig_dir, timesX_test=timesX_test, orig_lc_test=orig_lc_test,
                     objids_test=objids_test, passbands=passbands)



import os
import numpy as np
import pickle
from tensorflow.python.keras.models import Sequential
from tensorflow.python.keras.models import load_model
from tensorflow.python.keras.layers import Dense, Input
from tensorflow.python.keras.layers import LSTM, GRU
from tensorflow.python.keras.layers import Dropout, BatchNormalization, Activation, TimeDistributed, Masking
from tensorflow.python.keras.layers.convolutional import Conv1D, Conv2D
from tensorflow.python.keras.layers.convolutional import MaxPooling1D, MaxPooling2D
import matplotlib.pyplot as plt


def train_model(X_train, X_test, y_train, y_test, sample_weights=None, fig_dir='.', retrain=True, epochs=25, plot_loss=True):
    """ Train Neural Network classifier and save model. """

    model_filename = os.path.join(fig_dir, "keras_model.hdf5")
    #TODO: Try standard scaling and try normalising by peak and try reinputting such that it always normalises by largest value so far

    # colour = np.log10(X_train[:,:,0]) - np.log10(X_train[:,:,1])
    # X_train = np.dstack((X_train, colour))
    # colour = np.log10(X_test[:,:,0]) - np.log10(X_test[:,:,1])
    # X_test = np.dstack((X_test, colour))
    print("training...")
    if not retrain and os.path.isfile(model_filename):
        model = load_model(model_filename)
    else:
        num_classes = y_test.shape[-1]

        model = Sequential()

        model.add(Masking(mask_value=0.))

        # model.add(Conv1D(filters=32, kernel_size=3))
        # model.add(BatchNormalization())
        # model.add(Activation('relu'))
        # model.add(MaxPooling1D(pool_size=1))
        # model.add(Dropout(0.2, seed=42))

        model.add(LSTM(100, return_sequences=True))
        # model.add(Dropout(0.2, seed=42))
        # model.add(BatchNormalization())

        model.add(LSTM(100, return_sequences=True))
        # model.add(Dropout(0.2, seed=42))
        # model.add(BatchNormalization())
        # model.add(Dropout(0.2, seed=42))

        model.add(TimeDistributed(Dense(num_classes, activation='softmax')))
        model.compile(loss='categorical_crossentropy', optimizer='adam', metrics=['accuracy'])
        history = model.fit(X_train, y_train, validation_data=(X_test, y_test), epochs=epochs, batch_size=500,
                            verbose=2, sample_weight=sample_weights)

        print(model.summary())
        model.save(model_filename)

        with open(os.path.join(fig_dir, "model_history.pickle"), 'wb') as fp:
            pickle.dump(history.history, fp)

        if plot_loss:
            plot_history(history.history, fig_dir)

    return model


def plot_history(history, fig_dir):
    # Plot loss vs epochs
    plt.figure(figsize=(12,10))
    train_loss = history['loss']
    val_loss = history['val_loss']
    plt.plot(train_loss)
    plt.plot(val_loss)
    plt.ylabel('Loss')
    plt.xlabel('Epoch')
    plt.legend(['train', 'test'], loc='upper left')
    plt.title(os.path.basename(fig_dir).replace('_', '-'))
    plt.savefig(os.path.join(fig_dir, "model_loss_history.pdf"))

    # Plot accuracy vs figure
    plt.figure(figsize=(12,10))
    if 'accuracy' in history:
        train_acc = history['accuracy']
        val_acc = history['val_accuracy']
    else:
        train_acc = history['acc']
        val_acc = history['val_acc']
    plt.plot(train_acc)
    plt.plot(val_acc)
    plt.ylabel('Accuracy')
    plt.xlabel('Epoch')
    plt.legend(['train', 'test'], loc='upper left')
    plt.title(os.path.basename(fig_dir).replace('_', '-'))
    plt.savefig(os.path.join(fig_dir, "model_accuracy_history.pdf"))



import numpy as np
from collections import OrderedDict
import json
from six.moves.urllib.request import urlopen
from astroquery.irsa_dust import IrsaDust
import astropy.coordinates as coord
import astropy.units as u
import pickle
import urllib.request

from astrorapid.helpers import delete_indexes, convert_lists_to_arrays


def read_json(url):
    response = urlopen(url)
    return json.loads(response.read(), object_pairs_hook=OrderedDict)


def read_lasair_json(object_name='ZTF18acsovsw'):
    """
    Read light curve from lasair website API based on object name.

    Parameters
    ----------
    object_name : str
        The LASAIR object name. E.g. object_name='ZTF18acsovsw'

    """
    print(object_name)
    if isinstance(object_name, tuple):
        object_name, z_in = object_name
    else:
        z_in = None

    url = 'https://lasair.roe.ac.uk/object/{}/json/'.format(object_name)

    data = read_json(url)

    objid = data['objectId']
    ra = data['objectData']['ramean']
    dec = data['objectData']['decmean']
    # lasair_classification = data['objectData']['classification']
    tns_info = data['objectData']['annotation']
    photoZ = None
    for cross_match in data['crossmatches']:
        # print(cross_match)
        photoZ = cross_match['photoZ']
        separation_arcsec = cross_match['separationArcsec']
        catalogue_object_type = cross_match['catalogue_object_type']
    if z_in is not None and not np.isnan(z_in):
        redshift = z_in
    else:
        if photoZ is None:  # TODO: Get correct redshift
            try:
                if "z=" in tns_info:
                    photoZ = tns_info.split('z=')[1]
                    redshift = float(photoZ.replace(')', '').split()[0])
                    # print("PHOTOZZZZZZZZZZZZ", redshift, tns_info)
                elif "Z=" in tns_info:
                    photoZ = tns_info.split('Z=')[1]
                    redshift = float(photoZ.split()[0])
                    # print("PHOTOZZZZZZZZZZZZ", redshift, tns_info)
                else:
                    # return
                    print("TRYING ARBITRARY GUESS REDSHIFT = 0.1")
                    redshift = None
            except Exception as e:
                # return
                redshift = None
                print(e)
        else:
            redshift = photoZ

    print("Redshift is {}".format(redshift))
    if redshift is not None:
        objid += "_z={}".format(round(redshift, 2))

    # Get extinction  TODO: Maybe add this to RAPID code
    coo = coord.SkyCoord(ra * u.deg, dec * u.deg, frame='icrs')
    dust = IrsaDust.get_query_table(coo, section='ebv')
    mwebv = dust['ext SandF mean'][0]
    print("MWEBV")
    print(mwebv)

    mjd = []
    passband = []
    mag = []
    magerr = []
    photflag = []
    zeropoint = []
    dc_mag = []
    dc_magerr = []
    magnr, sigmagnr, isdiffpos = [], [], []
    for cand in data['candidates']:
        mjd.append(cand['mjd'])
        passband.append(cand['fid'])
        mag.append(cand['magpsf'])
        if 'sigmapsf' in cand:
            magerr.append(cand['sigmapsf'])
            photflag.append(4096)

            # if cand['magzpsci'] == 0:
            #     print("NO ZEROPOINT")
            #      zeropoint.append(26.2)  # TODO: Tell LASAIR their zeropoints are wrong
            # else:
            #     zeropoint.append(cand['magzpsci'])  #26.2) #
            zeropoint.append(cand['magzpsci'])
            dc_mag.append(cand['dc_mag'])
            dc_magerr.append(cand['dc_sigmag'])
            magnr.append(cand['magnr'])
            sigmagnr.append(cand['sigmagnr'])
            isdiffpos.append(cand['isdiffpos'])
        else:
            magerr.append(np.nan)#0.01 * cand['magpsf'])  #magerr.append(None)  #magerr.append(0.1 * cand['magpsf'])  #
            photflag.append(0)
            zeropoint.append(np.nan)#26.2)
            dc_mag.append(np.nan)
            dc_magerr.append(np.nan)
            magnr.append(np.nan)
            sigmagnr.append(np.nan)
            isdiffpos.append(None)

    mjd, passband, mag, magerr, photflag, zeropoint, dc_mag, dc_magerr, magnr, sigmagnr, isdiffpos = convert_lists_to_arrays(mjd, passband, mag, magerr, photflag, zeropoint, dc_mag, dc_magerr, magnr, sigmagnr, isdiffpos)

    # deleteindexes = np.where(magerr == None)  # [0,1,2,3,4,5,6,7,8,9,10,11,12,13,14,15,16,17,18,19,20,21,22,23,24,25,26,27,28,29]  #
    # mjd, passband, mag, magerr, photflag, zeropoint, dc_mag, dc_magerr, magnr, sigmagnr, isdiffpos = delete_indexes(deleteindexes, mjd, passband, mag, magerr, photflag, zeropoint, dc_mag, dc_magerr, magnr, sigmagnr, isdiffpos)
    deleteindexes = np.where((photflag==0) & (mjd > min(mjd[photflag>0])))  # delete where nondetections after first detection
    mjd, passband, mag, magerr, photflag, zeropoint, dc_mag, dc_magerr, magnr, sigmagnr, isdiffpos = delete_indexes(deleteindexes, mjd, passband, mag, magerr, photflag, zeropoint, dc_mag, dc_magerr, magnr, sigmagnr, isdiffpos)
    deleteindexes = np.where((mag < (np.median(mag[photflag==0]) - 0.5*np.std(mag[photflag==0]))) & (photflag==0)) # Remove non detection outliers
    mjd, passband, mag, magerr, photflag, zeropoint, dc_mag, dc_magerr, magnr, sigmagnr, isdiffpos = delete_indexes(deleteindexes, mjd, passband, mag, magerr, photflag, zeropoint, dc_mag, dc_magerr, magnr, sigmagnr, isdiffpos)

    return mjd, passband, mag, magerr, photflag, zeropoint, ra, dec, objid, redshift, mwebv, dc_mag, dc_magerr, magnr, sigmagnr, isdiffpos


def classify_lasair_light_curves(object_names=('ZTF18acsovsw',), savename=''):
    light_curve_list = []
    mjds, passbands, mags, magerrs, zeropoints, photflags = [], [], [], [], [], []
    dc_mags, dc_magerrs, magnrs, sigmagnrs, isdiffposs = [], [], [], [], []
    obj_names = []
    ras, decs, objids, redshifts, mwebvs = [], [], [], [], []
    peakmags_g, peakmags_r = [], []
    for object_name in object_names:
        try:
            mjd, passband, mag, magerr, photflag, zeropoint, ra, dec, objid, redshift, mwebv, dc_mag, dc_magerr, magnr, sigmagnr, isdiffpos = read_lasair_json(object_name)
            sortidx = np.argsort(mjd)
            mjds.append(mjd[sortidx])
            passbands.append(passband[sortidx])
            mags.append(mag[sortidx])
            magerrs.append(magerr[sortidx])
            zeropoints.append(zeropoint[sortidx])
            photflags.append(photflag[sortidx])
            dc_mags.append(dc_mag[sortidx])
            dc_magerrs.append(dc_magerr[sortidx])
            magnrs.append(magnr[sortidx])
            sigmagnrs.append(sigmagnr[sortidx])
            isdiffposs.append(isdiffpos[sortidx])

            obj_names.append(object_name)
            ras.append(ra)
            decs.append(dec)
            objids.append(objid)
            redshifts.append(redshift)
            mwebvs.append(mwebv)
            peakmags_g.append(min(mag[passband==1]))
            peakmags_r.append(min(mag[passband==2]))

        except Exception as e:
            print(e)
            continue

        flux = 10. ** (-0.4 * (mag - zeropoint))
        fluxerr = np.abs(flux * magerr * (np.log(10.) / 2.5))

        passband = np.where((passband == 1) | (passband == '1'), 'g', passband)
        passband = np.where((passband == 2) | (passband == '2'), 'r', passband)

        # Set photflag detections when S/N > 5
        photflag2 = np.zeros(len(flux))
        photflag2[flux / fluxerr > 5] = 4096
        photflag2[np.where(mjd == min(mjd[photflag2 == 4096]))] = 6144

        mjd_first_detection = min(mjd[photflag == 4096])
        photflag[np.where(mjd == mjd_first_detection)] = 6144

        deleteindexes = np.where(((passband == 3) | (passband == '3')) | (mjd > mjd_first_detection) & (photflag == 0))
        if deleteindexes[0].size > 0:
            print("Deleting indexes {} at mjd {} and passband {}".format(deleteindexes, mjd[deleteindexes], passband[deleteindexes]))
        mjd, passband, flux, fluxerr, zeropoint, photflag = delete_indexes(deleteindexes, mjd, passband, flux, fluxerr, zeropoint, photflag)

        light_curve_list += [(mjd, flux, fluxerr, passband, photflag, ra, dec, objid, redshift, mwebv)]

    with open(savename, 'wb') as f:
        pickle.dump([mjds, passbands, mags, magerrs, photflags, zeropoints, dc_mags, dc_magerrs, magnrs, sigmagnrs, isdiffposs, ras, decs, objids, redshifts, mwebvs], f)
    # np.savez('save_real_ZTF_unprocessed_data_snia_osc_12nov2019.npz', mjds=mjds, passbands=passbands, mags=mags, magerrs=magerrs, photflags=photflags, zeropoints=zeropoints, ras=ras, decs=decs, objids=objids, redshifts=redshifts, mwebvs=mwebvs)# , peakflux_g=peakfluxes_g, peakflux_r=peakfluxes_r)
    print("finished")


if __name__ == '__main__':
    # Get all sne as json
    json_url = 'https://api.sne.space/catalog?format=json'
    with urllib.request.urlopen(json_url) as url:
        data = json.loads(url.read().decode())

    # Check if ZTF alias
    ZTF_names = []
    ZTF_keys = []
    ZTF_redshifts = []
    ZTF_types = []

    SN = {
        'Ia': {},
        'Ia91T': {},
        'Ia91bg': {},
        'Iapec': {},
        'Iacsm': {},
        'Iax': {},
        'II': {},
        'IIP': {},
        'IIL': {},
        'IIpec': {},
        'IIn': {},
        'IIb': {},
        'Ib': {},
        'Ibn': {},
        'Ic': {},
        'IcBL': {},
        'Ibc': {},
        'SLSN': {},
        'SLSNI': {},
        'SLSNII': {},
        'CC': {}
    }


    for osc_name in data.keys():
        for alias in data[osc_name]['alias']:
            for name in alias.values():
                if name[0:3] == 'ZTF':
                    redshifts = np.array([entry['value'] for entry in data[osc_name]['redshift']]).astype('float')
                    claimed_types = [entry['value'] for entry in data[osc_name]['claimedtype']]
                    redshift = np.median(redshifts)
                    if len(claimed_types) == 0 or (len(claimed_types) == 1 and claimed_types[0] in ['Candidate', 'Other', 'other', 'removed', 'NT']):
                        continue

                    if len(claimed_types) == 1:
                        if claimed_types[0] in ['Ia',]:
                            SN['Ia'][name] = redshift
                        elif claimed_types[0] in ['Ia-91T']:
                            SN['Ia91T'][name] = redshift
                        elif claimed_types[0] in ['Ia-91bg']:
                            SN['Ia91bg'][name] = redshift
                        elif claimed_types[0] in ['Ia-02cx', 'Iax']:
                            SN['Iax'][name] = redshift
                        elif claimed_types[0] in ['Ia CSM']:
                            SN['Iacsm'][name] = redshift
                        elif claimed_types[0] in ['Ia Pec']:
                            SN['Iapec'][name] = redshift
                        elif claimed_types[0] in ['II']:
                            SN['II'][name] = redshift
                        elif claimed_types[0] in ['II P']:
                            SN['IIP'][name] = redshift
                        elif claimed_types[0] in ['II L']:
                            SN['IIL'][name] = redshift
                        elif claimed_types[0] in ['II Pec']:
                            SN['IIpec'][name] = redshift
                        elif claimed_types[0] in ['IIn']:
                            SN['IIn'][name] = redshift
                        elif claimed_types[0] in ['IIb', 'II/IIb']:
                            SN['IIb'][name] = redshift
                        elif claimed_types[0] in ['Ib']:
                            SN['Ib'][name] = redshift
                        elif claimed_types[0] in ['Ibn']:
                            SN['Ibn'][name] = redshift
                        elif claimed_types[0] in ['Ic']:
                            SN['Ic'][name] = redshift
                        elif claimed_types[0] in ['Ic BL']:
                            SN['IcBL'][name] = redshift
                        elif claimed_types[0] in ['Ib/c']:
                            SN['Ibc'][name] = redshift
                        elif claimed_types[0] in ['SLSN']:
                            SN['SLSN'][name] = redshift
                        elif claimed_types[0] in ['SLSN-I']:
                            SN['SLSNI'][name] = redshift
                        elif claimed_types[0] in ['SLSN-II']:
                            SN['SLSNII'][name] = redshift
                        else:
                            pass
                            # ZTF_names.append(name)
                            # ZTF_keys.append(osc_name)
                            # ZTF_redshifts.append(np.median(redshifts))
                            # ZTF_types.append(claimed_types)
                            # print(osc_name, name, redshifts, claimed_types)
                    elif len(claimed_types) == 2:
                        if 'Ia-91T' in claimed_types and 'Ia' in claimed_types:
                            SN['Ia91T'][name] = redshift
                        elif 'Ia-91bg' in claimed_types and 'Ia' in claimed_types:
                            SN['Ia91bg'][name] = redshift
                        elif 'Ia Pec' in claimed_types and 'Ia' in claimed_types:
                            SN['Iapec'][name] = redshift
                        elif 'Ia-99aa' in claimed_types and 'Ia' in claimed_types:
                            SN['Ia91T'][name] = redshift
                        elif 'IIn' in claimed_types and 'II' in claimed_types:
                            SN['IIn'][name] = redshift
                        elif 'II P' in claimed_types and 'II' in claimed_types:
                            SN['IIP'][name] = redshift
                        elif 'IIn' in claimed_types and 'II' in claimed_types:
                            SN['IIn'][name] = redshift
                        elif 'IIb' in claimed_types and 'II' in claimed_types:
                            SN['IIb'][name] = redshift
                        elif 'Ib' in claimed_types and 'CC' in claimed_types:
                            SN['Ib'][name] = redshift
                        elif 'II' in claimed_types and 'CC' in claimed_types:
                            SN['II'][name] = redshift
                        elif claimed_types[0][0:2] in ['II', 'Ib', 'Ic'] and claimed_types[1][0:2] in ['II', 'Ib', 'Ic']:
                            SN['CC'][name] = redshift
                        elif 'SLSN-I' in claimed_types and 'SLSN' in claimed_types:
                            SN['SLSNI'][name] = redshift
                        elif 'SLSN-II' in claimed_types and 'SLSN' in claimed_types:
                            SN['SLSNII'][name] = redshift
                        elif claimed_types[0][0:2] == 'Ia' and claimed_types[1][0:2] == 'Ia':
                            SN['Ia'][name] = redshift
                        else:
                            pass
                            # ZTF_names.append(name)
                            # ZTF_keys.append(osc_name)
                            # ZTF_redshifts.append(np.median(redshifts))
                            # ZTF_types.append(claimed_types)
                            # print(osc_name, name, redshifts, claimed_types)
                    else:
                        pass
                        # ZTF_names.append(name)
                        # ZTF_keys.append(osc_name)
                        # ZTF_redshifts.append(np.median(redshifts))
                        # ZTF_types.append(claimed_types)
                        # print(osc_name, name, redshifts, claimed_types)

                    ZTF_names.append(name)
                    ZTF_keys.append(osc_name)
                    ZTF_redshifts.append(np.median(redshifts))
                    ZTF_types.append(claimed_types)
                    print(osc_name, name, redshifts, claimed_types)

    for sntype in SN.keys():
        names_and_redshifts = list(SN[sntype].items())
        classify_lasair_light_curves(object_names=names_and_redshifts, savename=f'data/real_ZTF_data_from_osc/ZTF_data_{sntype}_osc-6-May-2020.pickle')




import pandas as pd
from copy import copy
import numpy as np
import matplotlib.pyplot as plt
import warnings
warnings.filterwarnings("ignore", category=UserWarning, module="matplotlib")

from astrorapid.custom_classifier import create_custom_classifier
from astrorapid.ANTARES_object.constants import TRIGGER_PHOTFLAG, GOOD_PHOTFLAG
from astrorapid.process_light_curves import InputLightCurve


BANDS = {'p48r': "r", 'p48g': "g"}
BANDS_VEC = np.vectorize(lambda x: BANDS[x])


def to_classnumber(lc):
    if lc.meta['classification'] == 'SN Ia':
        return 1
    else:
        return 2


def to_astrorapid(this_lc):
    if len(this_lc) < 2:
        return None

    this_lc = copy(this_lc)

    mask = np.logical_or.reduce((this_lc['band'] == 'p48r', this_lc['band'] == 'p48g'))
    this_lc = this_lc[mask]
    mask = this_lc['flux'] != 0.0
    this_lc = this_lc[mask]

    if len(this_lc) == 0:
        return None

    this_lc['band'] = BANDS_VEC(this_lc['band'])

    this_lc['photflag'] = [TRIGGER_PHOTFLAG] + (len(this_lc) - 1) * [GOOD_PHOTFLAG]

    class_number = to_classnumber(this_lc)

    return InputLightCurve(
        this_lc['mjd'],
        this_lc['flux'],
        this_lc['fluxerr'],
        this_lc['band'],
        this_lc['photflag'],
        this_lc.meta['ra'],
        this_lc.meta['dec'],
        this_lc.meta['ztfname'],
        redshift=this_lc.meta['z'],
        mwebv=this_lc.meta['mwebv'],
        training_set_parameters={'class_number': class_number, 'peakmjd': this_lc[np.argmax(this_lc['flux'])]['mjd']}
    )


def get_data(class_num, data_dir, save_dir, passbands, known_redshift, nprocesses, redo):
    lc_data_path = "rcf_marshallc_sncosmo_2018.pkl"
    lc_data = pd.read_pickle(lc_data_path)

    lightcurves = {str(to_classnumber(val)) + '_' + str(key): to_astrorapid(val) for key, val in lc_data.items()}
    valid_lightcurves = {key: val.preprocess_light_curve() for key, val in lightcurves.items() if val is not None}

    return valid_lightcurves


create_custom_classifier(get_data, 'data', plot=True, class_nums=(1,2), reread_data=False, retrain_network=False, nprocesses=1)




import os
import pickle
import numpy as np

from astrorapid.read_snana_fits import read_light_curves_from_snana_fits_files
from astrorapid.helpers import delete_indexes
from astrorapid.process_light_curves import InputLightCurve


def get_data(get_data_func, class_num, data_dir, save_dir, passbands, known_redshift=True, nprocesses=1, redo=False,
             calculate_t0=True):
    """
    Get data using some function.

    Parameters
    ----------
    get_data_func : func
        Function that returns light_curves and takes at least the arguments class_num, data_dir, save_dir, passbands.
        E.g. get_data_func = get_data_custom(class_num, data_dir, save_dir, passbands, known_redshift)
    class_num : int
        Class number. E.g. SNIa is 1. See helpers.py for lookup table.
        E.g. class_num = 1
    data_dir : str
        Directory where data is stored
        E.g. data_dir='data/ZTF_20190512/'
    save_dir : str
        Directory to save processed data
        E.g. save_dir='data/saved_light_curves/'
    passbands : tuple
        Passbands to use.
        E.g. passbands=('g', 'r')
    known_redshift : bool
        Whether to correct the light curves for cosmological time dilation or not.
    nprocesses : int or None
        Number of processes to use
    redo : bool
        Whether to redo reading the data and saving the processed data.
    calculate_t0 : bool
        Whether to calculate t0 during preprocessing.

    Returns
    -------
        light_curves : dict of astropy.table.Table objects
            e.g light_curves['objid1'] =
                passband   time       flux     fluxErr   photflag
                  str1   float32    float32    float32    int32
                -------- -------- ----------- ---------- --------
                       g -46.8942  -48.926975  42.277767        0
                       g -43.9352  -105.35379   72.97575        0
                       g -35.9161  -46.264206    99.9172        0
                       g -28.9377  -28.978344  42.417065        0
                       g -25.9787  109.886566   46.03949        0
                       g -15.0399    -80.2485   80.38155        0
                       g -12.0218    93.51743  113.21529        0
                       g  -6.9585   248.88364 108.606865        0
                       g  -4.0411   341.41498  47.765404        0
                       g      0.0    501.7441   45.37485     6144
                     ...      ...         ...        ...      ...
                       r  40.9147   194.32494  57.836903     4096
                       r  59.9162    67.59185   45.66463     4096
                       r  62.8976    80.85155  44.356197     4096
                       r  65.8974   28.174305   44.75049     4096
                       r  71.8966  -18.790287 108.049774     4096
                       r  74.9297  -3.1707647  125.15057     4096
                       r  77.9341 -11.0205965 125.784676     4096
                       r  80.8576   129.65466   69.99305     4096
                       r  88.8922  -14.259436  52.917866     4096
                       r 103.8734   27.178356 115.537704     4096
    """

    return get_data_func(class_num, data_dir, save_dir, passbands, known_redshift, nprocesses, redo)


def get_data_from_snana_fits(class_num, data_dir='data/ZTF_20190512/', save_dir='data/saved_light_curves/',
                             passbands=('g', 'r'), known_redshift=True, nprocesses=1, redo=False, calculate_t0=True):
    """
    Get data from SNANA fits data files.

    """
    save_lc_filepath = os.path.join(save_dir, f"lc_classnum_{class_num}.pickle")

    if os.path.exists(save_lc_filepath) and not redo:
        with open(save_lc_filepath, "rb") as fp:  # Unpickling
            light_curves = pickle.load(fp)
    else:
        class_dir = os.path.join(data_dir, 'ZTF_MSIP_MODEL{:02d}'.format(class_num))
        files = os.listdir(class_dir)

        head_files = []
        phot_files = []
        for file in files:
            filepath = os.path.join(data_dir, class_dir, file)
            if filepath.endswith('HEAD.FITS'):
                head_files.append(filepath)
                phot_files.append(filepath.replace('_HEAD.FITS', '_PHOT.FITS'))
            print(filepath)

        light_curves = read_light_curves_from_snana_fits_files(head_files, phot_files, passbands,
                                                               known_redshift=known_redshift, nprocesses=nprocesses,
                                                               calculate_t0=calculate_t0)

        with open(save_lc_filepath, "wb") as fp:  # Pickling
            pickle.dump(light_curves, fp)

    return light_curves


def get_real_ztf_training_data(class_name, data_dir='data/real_ZTF_data_from_osc',
                               save_dir='data/saved_light_curves/', pbs=('g', 'r'),
                               known_redshift=True, nprocesses=1, redo=False, calculate_t0=True):
    """
    Get data from saved real ZTF data with names and types from the Open Supernova Catalog
    """

    save_lc_filepath = os.path.join(save_dir, f"lc_classnum_{class_name}.pickle")
    if os.path.exists(save_lc_filepath) and not redo:
        with open(save_lc_filepath, "rb") as fp:  # Unpickling
            light_curves = pickle.load(fp)
    else:
        light_curves = {}
        data_filepath = os.path.join(data_dir, f"ZTF_data_{class_name}_osc-5-Apr-2020.pickle")
        with open(data_filepath, "rb") as fp:
            mjds, passbands, mags, magerrs, photflags, zeropoints, dc_mags, dc_magerrs, magnrs, \
            sigmagnrs, isdiffposs, ras, decs, objids, redshifts, mwebvs = pickle.load(fp)

        for i, objid in enumerate(objids):
            if known_redshift and (redshifts[i] is None or np.isnan(redshifts[i])):
                print(f"Skipping {objid} because redshift is unknown and known_redshift model is selected")
                continue

            flux = 10. ** (-0.4 * (mags[i] - zeropoints[i]))
            fluxerr = np.abs(flux * magerrs[i] * (np.log(10.) / 2.5))

            passbands[i] = np.where((passbands[i] == 1) | (passbands[i] == '1'), 'g', passbands[i])
            passbands[i] = np.where((passbands[i] == 2) | (passbands[i] == '2'), 'r', passbands[i])

            mjd_first_detection = min(mjds[i][photflags[i] == 4096])
            photflags[i][np.where(mjds[i] == mjd_first_detection)] = 6144

            deleteindexes = np.where(((passbands[i] == 3) | (passbands[i] == '3')) | ((mjds[i] > mjd_first_detection) & (photflags[i] == 0)) | (np.isnan(flux)))
            if deleteindexes[0].size > 0:
                print("Deleting indexes {} at mjd {} and passband {}".format(deleteindexes, mjds[i][deleteindexes], passbands[i][deleteindexes]))
            mjd, passband, flux, fluxerr, zeropoint, photflag = delete_indexes(deleteindexes, mjds[i], passbands[i], flux, fluxerr, zeropoints[i], photflags[i])
            peakmjd = mjd[np.argmax(flux)]

            inputlightcurve = InputLightCurve(mjd, flux, fluxerr, passband, photflag,
                                              ras[i], decs[i], objid, redshifts[i], mwebvs[i],
                                              known_redshift=known_redshift,
                                              training_set_parameters={'class_number': class_name,
                                                                       'peakmjd': peakmjd},
                                              calculate_t0=calculate_t0)
            light_curves[objid] = inputlightcurve.preprocess_light_curve()

        with open(save_lc_filepath, "wb") as fp:
            pickle.dump(light_curves, fp)

    return light_curves



import os
import numpy as np
import pandas as pd
from distutils.spawn import find_executable
from scipy.interpolate import interp1d
from tensorflow.keras.utils import to_categorical
from sklearn.metrics import confusion_matrix
from collections import OrderedDict

from astrorapid.classifier_metrics import plot_confusion_matrix, compute_multiclass_roc_auc, compute_precision_recall, \
    plasticc_log_loss

try:
    import matplotlib.pyplot as plt
    import matplotlib
    from matplotlib.ticker import MaxNLocator
    import matplotlib.animation as animation
    import imageio
    if find_executable('latex'):
        plt.rcParams['text.usetex'] = True
    plt.rcParams['font.serif'] = ['Computer Modern Roman'] + plt.rcParams['font.serif']

except ImportError:
    print("Warning: You will need to install 'matplotlib' and 'imageio' if you want to plot the "
          "classification performance metrics.")

COLORS = ['grey', 'tab:green', 'tab:orange', 'tab:blue', 'tab:red', 'tab:purple', 'tab:brown', '#aaffc3', 'tab:olive',
          'tab:cyan', '#FF1493', 'navy', 'tab:pink', 'lightcoral', '#228B22', '#aa6e28', '#FFA07A']
COLCLASS = {'Pre-explosion': 'grey', 'SNIa-norm': 'tab:green', 'SNIbc': 'tab:orange', 'SNII': 'tab:blue',
            'SNIa-91bg': 'tab:red', 'SNIa-x': 'tab:purple', 'point-Ia': 'tab:brown', 'Kilonova': '#aaffc3',
            'SLSN-I': 'tab:olive', 'PISN': 'tab:cyan', 'ILOT': '#FF1493', 'CART': 'navy', 'TDE': 'tab:pink',
            'AGN': 'bisque'}
COLPB = {'u': 'tab:blue', 'g': 'tab:blue', 'r': 'tab:orange', 'i': 'm', 'z': 'k', 'Y': 'y'}
MARKPB = {'g': 'o', 'r': 's'}
ALPHAPB = {'g': 0.3, 'r': 1.}
WLOGLOSS_WEIGHTS = [1, 1, 1, 1, 1, 1, 1, 2, 2, 2, 2, 2, 2]
MINTIME = -70
MAXTIME = 80


def plot_metrics(class_names, model, X_test, y_test, fig_dir, timesX_test=None, orig_lc_test=None, objids_test=None,
                 passbands=('g', 'r')):
    scores = model.evaluate(X_test, y_test, verbose=0)
    print("Accuracy: %.2f%%" % (scores[1] * 100))

    class_names = sorted(class_names)

    y_pred = model.predict(X_test)
    y_test_indexes = np.argmax(y_test, axis=-1)
    y_pred_indexes = np.argmax(y_pred, axis=-1)

    accuracy = len(np.where(y_pred_indexes == y_test_indexes)[0])
    print("Accuracy is: {}/{} = {}".format(accuracy, len(y_test_indexes.flatten()),
                                           accuracy / len(y_test_indexes.flatten())))

    class_names = ["Pre-explosion"] + class_names

    # Set trailing zeros to -200
    for i in range(timesX_test.shape[0]):
        timesX_test[i][:np.argmin(timesX_test[i])] = -200
        timesX_test[i][np.argmax(timesX_test[i]) + 1:] = -200

    for cname in class_names:
        dirname = os.path.join(fig_dir + '/lc_pred', cname)
        if not os.path.exists(dirname):
            os.makedirs(dirname)

    # Plot accuracy vs time per class
    font = {'family': 'normal',
            'size': 36}
    matplotlib.rc('font', **font)

    # Plot classification example vs time
    for idx in np.arange(0, 100):
        true_class = int(max(y_test_indexes[idx]))
        print(true_class)
        # if true_class != 1:
        #     continue
        print("Plotting example vs time", idx, objids_test[idx])
        argmax = timesX_test[idx].argmax() + 1

        lc_data = orig_lc_test[idx]
        used_passbands = [pb for pb in passbands if pb in lc_data['passband']]

        new_t = np.concatenate([lc_data[lc_data['passband'] == pb]['time'].data for pb in used_passbands])
        new_t = np.sort(new_t[~np.isnan(new_t)])
        new_y_predict = []

        fig, (ax1, ax2) = plt.subplots(nrows=2, ncols=1, figsize=(13, 15), num="classification_vs_time_{}".format(idx),
                                       sharex=True)

        for pbidx, pb in enumerate(passbands):
            if pb not in used_passbands:
                continue
            pbmask = lc_data['passband'] == pb
            # masktime = (lc_data[pbmask]['time'] > MINTIME) & (lc_data[pbmask]['time'] < MAXTIME)
            ax1.errorbar(lc_data[pbmask]['time'], lc_data[pbmask]['flux'],
                         yerr=lc_data[pbmask]['fluxErr'], fmt=MARKPB[pb], label=pb, c=COLPB[pb],
                         lw=3, markersize=10, alpha=0.2)
            ax1.plot(timesX_test[idx][:argmax], X_test[idx][:, pbidx][:argmax], c=COLPB[pb],
                     lw=3)  # , markersize=10, marker=MARKPB[pb])

        true_class = int(max(y_test_indexes[idx]))
        ax1.axvline(x=0, color='k', linestyle='-', linewidth=1)
        ax2.axvline(x=0, color='k', linestyle='-', linewidth=1)
        try:
            redshift = lc_data.meta['redshift']
            mwebv = lc_data.meta['mwebv']
            b = lc_data.meta['b']
            trigger_mjd = lc_data.meta['trigger_mjd']
            t0 = lc_data.meta['t0']
            peakmjd = lc_data.meta['peakmjd']

            if t0 != -99:
                ax1.axvline(x=t0, color='grey', linestyle='--', linewidth=2)
                ax2.axvline(x=t0, color='grey', linestyle='--', linewidth=2)
                ax1.annotate('$t_0 = {}$'.format(round(t0, 1)), xy=(t0, 1), xytext=(t0 - 33, 0.9*max(orig_lc_test[idx]['r']['flux'])), color='grey')

            ax1.axvline(x=peakmjd - trigger_mjd, color='k', linestyle=':', linewidth=1)
            ax2.axvline(x=peakmjd - trigger_mjd, color='k', linestyle=':', linewidth=1)
        except Exception as e:
            print(e)

        class_accuracies = [timesX_test[idx][:argmax]]

        #
        for classnum, classname in enumerate(class_names):
            new_y_predict.append(np.interp(new_t, timesX_test[idx][:argmax], y_pred[idx][:, classnum][:argmax]))

        for classnum, classname in enumerate(class_names):
            ax2.plot(timesX_test[idx][:argmax], y_pred[idx][:, classnum][:argmax], '-', label=classname,
                     color=COLORS[classnum], linewidth=3)
            # ax2.step(new_t, new_y_predict[classnum], '-', label=classname,
            #          color=COLORS[classnum], linewidth=3, where='post')
            class_accuracies.append(y_pred[idx][:, classnum][:argmax])
        ax1.legend(frameon=True, fontsize=33)
        ax2.legend(frameon=True, fontsize=20.5, ncol=1, loc='right')
        ax2.set_xlabel("Days since trigger (rest frame)")  # , fontsize=18)
        ax1.set_ylabel("Relative Flux")  # , fontsize=15)
        ax2.set_ylabel("Class Probability")  # , fontsize=18)
        # ax1.set_ylim(-0.1, 1.1)
        # ax2.set_ylim(0, 1)
        mintime_lc = min([min(lc_data[lc_data['passband'] == pb]['time']) for pb in used_passbands])
        maxtime_lc = max([max(lc_data[lc_data['passband'] == pb]['time']) for pb in used_passbands])
        ax1.set_xlim(max(mintime_lc, MINTIME), min(maxtime_lc, MAXTIME))
        # ax1.set_xlim(MINTIME, MAXTIME)
        plt.setp(ax1.get_xticklabels(), visible=False)
        ax2.yaxis.set_major_locator(MaxNLocator(nbins=6, prune='upper'))  # added
        plt.tight_layout()
        fig.subplots_adjust(hspace=0)
        plt.savefig(os.path.join(fig_dir + '/lc_pred',
                                 "{}_{}_{}_{}_{}_matrix_input2.pdf".format(objids_test[idx], idx, class_names[true_class], redshift,
                                                                                 peakmjd - trigger_mjd)))
        plt.savefig(os.path.join(fig_dir + '/lc_pred', class_names[true_class],
                                 "{}_{}_{}_{}_{}_matrix_input2.pdf".format(objids_test[idx], idx, class_names[true_class], redshift,
                                                                                 peakmjd - trigger_mjd)))
        plt.close()

    # Plot animated classification example vs time
    for idx in []: #[869]: #[181, 409, 491, 508, 765, 1156, 1335, 1358, 1570]:  # np.arange(765, 766):  # [0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 20, 25, 30]:
        print("Plotting animation example vs time", idx)

        #
        new_t = np.concatenate([lc_data[lc_data['passband'] == pb]['time'].values for pb in used_passbands])
        new_t = np.sort(new_t[~np.isnan(new_t)])
        new_y_predict = []
        # all_flux = list(orig_lc_test[idx]['g']['flux']) + list(orig_lc_test[idx]['r']['flux'])

        # timestep = timesX_test[idx][1] - timesX_test[idx][0]
        argmax = timesX_test[idx].argmax() + 1
        fig, (ax1, ax2) = plt.subplots(nrows=2, ncols=1, figsize=(13, 15), num="animation_classification_vs_time_{}".format(idx), sharex=True)

        ax1.legend(frameon=True, fontsize=33)
        ax2.legend(frameon=True, fontsize=20, loc='center right')  # , ncol=2)
        ax2.set_xlabel("Days since trigger (rest frame)")  # , fontsize=18)
        ax1.set_ylabel("Relative Flux")  # , fontsize=15)
        ax2.set_ylabel("Class Probability")  # , fontsize=18)
        ax2.set_ylim(bottom=0, top=0.9)
        ax1.set_ylim(top=1.15*max([max(lc_data[lc_data['passband'] == pb]['flux']) for pb in used_passbands]), bottom=-2700)
        ax1.set_xlim(MINTIME, MAXTIME)
        plt.setp(ax1.get_xticklabels(), visible=False)
        ax2.yaxis.set_major_locator(MaxNLocator(nbins=6, prune='upper'))  # added
        plt.tight_layout()
        fig.subplots_adjust(hspace=0)

        true_class = int(max(y_test_indexes[idx]))

        redshift = lc_data.meta['redshift']
        mwebv = lc_data.meta['mwebv']
        b = lc_data.meta['b']
        trigger_mjd = lc_data.meta['trigger_mjd']
        t0 = lc_data.meta['t0']
        peakmjd = lc_data.meta['peakmjd']

        t0 = -7.3
        ax1.axvline(x=t0, color='grey', linestyle='--', linewidth=2)
        ax2.axvline(x=t0, color='grey', linestyle='--', linewidth=2)
        ax1.axvline(x=0, color='k', linestyle='-', linewidth=1)
        ax2.axvline(x=0, color='k', linestyle='-', linewidth=1)
        # ax1.axvline(x=peakmjd - trigger_mjd, color='k', linestyle=':', linewidth=1)
        # ax2.axvline(x=peakmjd - trigger_mjd, color='k', linestyle=':', linewidth=1)
        ax1.annotate('$t_0 = {}$'.format(round(t0, 1)), xy=(t0, 1), xytext=(t0 - 30, 0.9*max(orig_lc_test[idx]['r']['flux'])), color='grey')

        Writer = animation.writers['ffmpeg']
        writer = Writer(fps=5, bitrate=1800)

        #
        for classnum, classname in enumerate(class_names):
            new_y_predict.append(np.interp(new_t, timesX_test[idx][:argmax], y_pred[idx][:, classnum][:argmax]))

        def animate(i):
            for pbidx, pb in enumerate(passbands):
                if pb not in used_passbands:
                    continue
                # ax1.plot(timesX_test[idx][:argmax][:int(i+1)], X_test[idx][:, pbidx][:argmax][:int(i+1)], label=pb, c=COLPB[pb], lw=3)#, markersize=10, marker=MARKPB[pb])
                if i + 1 >= len(new_t):
                    break
                # If less than 0.4 day gap in times skip
                if (i + 1) < len(new_t) and (new_t[int(i + 1)] - new_t[int(i)]) < 0.4:
                    break

                dea = [lc_data[pbmask]['time'] < new_t[int(i + 1)]]

                ax1.errorbar(np.array(lc_data[pbmask]['time'])[dea], np.array(lc_data[pbmask]['flux'])[dea],
                             yerr=np.array(lc_data[pbmask]['fluxErr'])[dea], fmt=MARKPB[pb], label=pb,
                             c=COLPB[pb], lw=3, markersize=10)

            for classnum, classname in enumerate(class_names):
                # ax2.plot(timesX_test[idx][:argmax][:int(i+1)], y_pred[idx][:, classnum][:argmax][:int(i+1)], '-', label=classname, color=COLORS[classnum], linewidth=3)
                ax2.step(new_t[:int(i + 1)], new_y_predict[classnum][:int(i + 1)], '-', label=classname,
                         color=COLCLASS[classname], linewidth=3, where='post')

            # Don't repeat legend items
            ax1.legend(frameon=True, fontsize=33)
            ax2.legend(frameon=True, fontsize=20, loc='center right')
            handles1, labels1 = ax1.get_legend_handles_labels()
            handles2, labels2 = ax2.get_legend_handles_labels()
            by_label1 = OrderedDict(zip(labels1, handles1))
            by_label2 = OrderedDict(zip(labels2, handles2))
            ax1.legend(by_label1.values(), by_label1.keys(), frameon=True, fontsize=33)
            ax2.legend(by_label2.values(), by_label2.keys(), frameon=True, fontsize=20, loc='center right')

        ani = animation.FuncAnimation(fig, animate, frames=len(new_t), repeat=True)
        ani.save(os.path.join(fig_dir + '/lc_pred', "classification_vs_time_{}_{}_{}_{}.mp4".format(idx, class_names[true_class], redshift, peakmjd - trigger_mjd)), writer=writer)

    print("Plotting Accuracy vs time per class...")
    fig = plt.figure("accuracy_vs_time_perclass", figsize=(13, 12))
    # fig = plt.figure(figsize=(13, 12))
    for classnum, classname in enumerate(class_names):
        correct_predictions_inclass = (y_test_indexes == classnum) & (y_pred_indexes == y_test_indexes)
        time_bins = np.arange(-150, 150, 3.)

        times_binned_indexes = np.digitize(timesX_test, bins=time_bins, right=True)
        time_list_indexes_inclass, count_correct_vs_binned_time_inclass = np.unique(times_binned_indexes * correct_predictions_inclass, return_counts=True)
        time_list_indexes2_inclass, count_objects_vs_binned_time_inclass = np.unique(times_binned_indexes * (y_test_indexes == classnum), return_counts=True)

        time_list_indexes_inclass = time_list_indexes_inclass[time_list_indexes_inclass < len(time_bins)]
        count_correct_vs_binned_time_inclass = count_correct_vs_binned_time_inclass[time_list_indexes_inclass < len(time_bins)]
        time_list_indexes2_inclass = time_list_indexes2_inclass[time_list_indexes2_inclass < len(time_bins)]
        count_objects_vs_binned_time_inclass = count_objects_vs_binned_time_inclass[time_list_indexes2_inclass < len(time_bins)]

        start_time_index = int(np.where(time_list_indexes2_inclass == time_list_indexes_inclass[0])[0])
        end_time_index = int(np.where(time_list_indexes2_inclass == time_list_indexes_inclass[-1])[0]) + 1

        try:
            accuracy_vs_time_inclass = count_correct_vs_binned_time_inclass[1:] / count_objects_vs_binned_time_inclass[start_time_index:end_time_index]
        except Exception as e:
            print(e)
            continue

        try:
            assert np.all(time_list_indexes_inclass[1:] == time_list_indexes2_inclass[start_time_index:end_time_index])
        except Exception as e:
            import pdb; pdb.set_trace()
            print(e)
            pass

        plt.plot(time_bins[time_list_indexes_inclass[1:]], accuracy_vs_time_inclass, '-', label=classname, color=COLORS[classnum], lw=3)
    plt.xlim(left=-35, right=70)
    plt.xlabel("Days since trigger (rest frame)")
    plt.ylabel("Classification accuracy")
    plt.legend(frameon=True, fontsize=25, ncol=2, loc=0)  # , bbox_to_anchor=(0.05, -0.1))
    plt.tight_layout()
    plt.savefig(os.path.join(fig_dir, "accuracy_vs_time_perclass.pdf"), bbox_inches='tight')
    plt.close()

    font = {'family': 'normal',
            'size': 33}

    matplotlib.rc('font', **font)

    # Plot confusion matrix at different days past trigger
    print("Plotting Confusion Matrices...")
    time_bins = np.arange(-110, 110, 1.)
    nobjects = len(timesX_test)
    ntimesteps = len(time_bins)
    nclasses = y_test.shape[-1]
    y_test_indexes_days_past_trigger = np.zeros((nobjects, ntimesteps))
    y_pred_indexes_days_past_trigger = np.zeros((nobjects, ntimesteps))
    y_pred_days_past_trigger = np.zeros((nobjects, ntimesteps, nclasses))
    for objidx in range(nobjects):
        print(objidx, nobjects, 'For conf matrix') if objidx % 1000 == 0 else 0
        f = interp1d(timesX_test[objidx], y_test_indexes[objidx], kind='nearest', bounds_error=False,
                     fill_value='extrapolate')
        y_test_indexes_days_past_trigger[objidx][:] = f(time_bins)
        f = interp1d(timesX_test[objidx], y_pred_indexes[objidx], kind='nearest', bounds_error=False,
                     fill_value='extrapolate')
        y_pred_indexes_days_past_trigger[objidx][:] = f(time_bins)
        for classidx in range(nclasses):
            classprob = y_pred[objidx][:, classidx]
            mintimeidx = 0
            maxtimeidx = np.argmax(timesX_test[objidx])
            f = interp1d(timesX_test[objidx], classprob, kind='linear', bounds_error=False,
                         fill_value=(classprob[mintimeidx], classprob[maxtimeidx]))
            y_pred_days_past_trigger[objidx][:, classidx][:] = f(time_bins)

    images_cf, images_roc, images_pr = [], [], []
    roc_auc, pr_auc = {}, {}
    wlogloss = {}
    for i, days_since_trigger in enumerate(list(np.arange(0, 75,step=2))):# + list(np.arange(25, 70, step=5))):  # [('early2days', 2), ('late40days', 40)]: # [-25, -20, -15, -10, -5, 0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 15, 20, 25, 30, 35, 40, 45, 50, 55, 60, 65, 70]:
        print("Plotting CF matrix", i, days_since_trigger, "days")
        index = np.where(time_bins == days_since_trigger)[0][0]
        y_test_on_day_i = y_test_indexes_days_past_trigger[:, index]
        y_pred_on_day_i = y_pred_indexes_days_past_trigger[:, index]
        y_pred_prob_on_day_i = y_pred_days_past_trigger[:, index]

        name = 'since_trigger_{}'.format(days_since_trigger)
        title = '{} days since trigger'.format(days_since_trigger)
        cnf_matrix = confusion_matrix(y_test_on_day_i, y_pred_on_day_i)
        print(name, cnf_matrix)
        figname_cf = plot_confusion_matrix(cnf_matrix, classes=class_names[1:], normalize=True, title=title,
                                           fig_dir=fig_dir + '/cf_since_trigger', name=name)
        images_cf.append(imageio.imread(figname_cf))

        figname_roc, roc_auc[days_since_trigger] = compute_multiclass_roc_auc(class_names,
                                                                              to_categorical(y_test_on_day_i,
                                                                                             num_classes=nclasses),
                                                                              y_pred_prob_on_day_i, name=name,
                                                                              fig_dir=fig_dir + '/roc_since_trigger',
                                                                              title=title)
        images_roc.append(imageio.imread(figname_roc))

        figname_pr, pr_auc[days_since_trigger] = compute_precision_recall(class_names, to_categorical(y_test_on_day_i,
                                                                                                      num_classes=nclasses),
                                                                          y_pred_prob_on_day_i, name=name,
                                                                          fig_dir=fig_dir + '/pr_since_trigger',
                                                                          title=title)
        images_pr.append(imageio.imread(figname_pr))

        try:
            wlogloss[days_since_trigger] = plasticc_log_loss(to_categorical(y_test_on_day_i, num_classes=nclasses),
                                                             y_pred_prob_on_day_i, relative_class_weights=WLOGLOSS_WEIGHTS)
        except Exception as e:
            print("Cannot compute weighted PLAsTiCC Log Loss.", e)

        plt.close()

        objids_filename = os.path.join(fig_dir + '/truth_table_since_trigger', 'objids_{}.csv'.format(name))
        predicted_table_filename = os.path.join(fig_dir + '/truth_table_since_trigger',
                                                'predicted_prob_{}.csv'.format(name))
        truth_table_filename = os.path.join(fig_dir + '/truth_table_since_trigger', 'truth_table_{}.csv'.format(name))
        np.savetxt(objids_filename, objids_test, fmt='%s')
        np.savetxt(predicted_table_filename, y_pred_prob_on_day_i)
        np.savetxt(truth_table_filename, to_categorical(y_test_on_day_i, num_classes=nclasses))
        # out_objtable = os.path.join(fig_dir + '/truth_table_since_trigger', 'out_obj_table_{}.txt'.format(name))
        # out_truth = os.path.join(fig_dir + '/truth_table_since_trigger', 'out_truth_{}.csv'.format(name))
        # out_pred = os.path.join(fig_dir + '/truth_table_since_trigger', 'out_pred_{}.csv'.format(name))
        # make_tables(objids_filename, predicted_table_filename, truth_table_filename, directory='', processes=2, out_objtable=out_objtable, out_truth=out_truth, out_pred=out_pred)

    imageio.mimsave(os.path.join(fig_dir, 'animation_cf_since_trigger.gif'), images_cf, duration=0.25)
    imageio.mimsave(os.path.join(fig_dir, 'animation_roc_since_trigger.gif'), images_roc, duration=0.25)
    imageio.mimsave(os.path.join(fig_dir, 'animation_pr_since_trigger.gif'), images_pr, duration=0.25)

    font = {'family': 'normal',
            'size': 36}

    matplotlib.rc('font', **font)

    # Plot ROC AUC vs time
    plt.figure("ROC AUC vs time", figsize=(13, 12))
    roc_auc = pd.DataFrame(roc_auc).transpose()
    names = list(roc_auc.keys())

    plt.plot(list(roc_auc['micro'].index), list(roc_auc['micro'].values), color='navy', linestyle=':', linewidth=4,
             label='micro-average')
    for classnum, classname in enumerate(class_names):
        if classname not in names or classnum == 0:
            print(classname)
            continue
        times = list(roc_auc[classname].index)
        aucs = list(roc_auc[classname].values)
        plt.plot(times, aucs, '-', label=classname, color=COLORS[classnum], linewidth=4)

    plt.legend(frameon=True, fontsize=25)
    plt.xlabel("Days since trigger (rest frame)")  # , fontsize=18)
    plt.ylabel("AUC")  # , fontsize=15)
    plt.tight_layout()
    plt.savefig(os.path.join(fig_dir, "auc_roc_vs_time.pdf"))
    plt.close()

    # # Plot Precision AUC vs time
    # plt.figure("ROC AUC vs time", figsize=(12, 11))
    # pr_auc = pd.DataFrame(pr_auc).transpose()
    # names = list(pr_auc.keys())
    # plt.plot(list(pr_auc['micro'].index), list(pr_auc['micro'].values), color='navy', linestyle=':', linewidth=4, label='micro')
    # for classnum, classname in enumerate(class_names):
    #     if classname not in names or classnum == 0:
    #         print(classname)
    #         continue
    #     times = list(pr_auc[classname].index)
    #     aucs = list(pr_auc[classname].values)
    #     plt.plot(times, aucs, '-', label=classname, color=COLORS[classnum], linewidth=2)
    #
    # plt.legend(frameon=False, fontsize=19)
    # plt.xlabel("Days since trigger")  # , fontsize=18)
    # plt.ylabel("AUC")  # , fontsize=15)
    # plt.savefig(os.path.join(fig_dir, "auc_pr_vs_time.pdf"))
    # plt.close()

    font = {'family': 'normal',
            'size': 36}
    matplotlib.rc('font', **font)

    # Plot weighted log loss vs time
    plt.figure("Weighted log loss vs time", figsize=(13, 12))
    plt.plot(list(wlogloss.keys()), list(wlogloss.values()), linewidth=4, label='Weighted Log loss')
    plt.xlabel("Days since trigger (rest frame)")  # , fontsize=18)
    plt.ylabel("Weighted log loss")  # , fontsize=15)
    plt.tight_layout()
    plt.savefig(os.path.join(fig_dir, "wlogloss_vs_time.pdf"))
    plt.close()
    print(wlogloss)




    # for idx in [742]:#np.arange(0, 1000):
    #     true_class = int(max(y_test_indexes[idx]))
    #     print(true_class)
    #     if true_class != 1:
    #         continue
    #     print("Plotting example vs time", idx)
    #     argmax = timesX_test[idx].argmax() + 1
    #
    #     #
    #     new_t = np.array([orig_lc_test[idx][pb]['time'].values for pb in passbands]).flatten()
    #     new_t = np.sort(new_t[~np.isnan(new_t)])
    #     new_y_predict = []
    #
    #     fig, (ax1, ax2) = plt.subplots(nrows=2, ncols=1, figsize=(13, 15), num="classification_vs_time_{}".format(idx),
    #                                    sharex=True)
    #
    #     for pb in passbands:
    #         if pb in orig_lc_test[idx].keys():
    #             try:
    #                 ax1.errorbar(lc_data[pbmask]['time'], lc_data[pbmask]['flux'],
    #                              yerr=orig_lc_test[idx][pb]['fluxErr'], fmt=MARKPB[pb], label=pb, c=COLPB[pb],
    #                              lw=3, markersize=10)
    #             except KeyError:
    #                 ax1.errorbar(orig_lc_test[idx][pb]['time'], orig_lc_test[idx][pb][5], yerr=orig_lc_test[idx][pb][6],
    #                              fmt=MARKPB[pb], label=pb, c=COLPB[pb], lw=3, markersize=10)
    #     true_class = int(max(y_test_indexes[idx]))
    #     ax1.axvline(x=0, color='k', linestyle='-', linewidth=1)
    #     ax2.axvline(x=0, color='k', linestyle='-', linewidth=1)
    #     try:
    #         otherinfo = orig_lc_test[idx]['otherinfo'].values.flatten()
    #         redshift, b, mwebv, trigger_mjd, t0, peakmjd = otherinfo[0:6]
    #         t0=-7.3
    #         ax1.axvline(x=t0, color='grey', linestyle='--', linewidth=2)
    #         ax2.axvline(x=t0, color='grey', linestyle='--', linewidth=2)
    #         ax1.annotate('$t_0 = {}$'.format(round(t0, 1)), xy=(t0, 1), xytext=(t0 - 33, 0.9*max(orig_lc_test[idx]['r']['flux'])), color='grey')
    #         print(otherinfo[0:6])
    #         # ax1.axvline(x=peakmjd - trigger_mjd, color='k', linestyle=':', linewidth=1)
    #         # ax2.axvline(x=peakmjd - trigger_mjd, color='k', linestyle=':', linewidth=1)
    #     except Exception as e:
    #         print(e)
    #
    #     class_accuracies = [timesX_test[idx][:argmax]]
    #
    #     #
    #     for classnum, classname in enumerate(class_names):
    #         new_y_predict.append(np.interp(new_t, timesX_test[idx][:argmax], y_pred[idx][:, classnum][:argmax]))
    #
    #     for classnum, classname in enumerate(class_names):
    #         # ax2.plot(timesX_test[idx][:argmax], y_pred[idx][:, classnum][:argmax], '-', label=classname,
    #         #          color=COLORS[classnum], linewidth=3)
    #         ax2.step(new_t, new_y_predict[classnum], '-', label=classname,
    #                  color=COLORS[classnum], linewidth=3, where='post')
    #         class_accuracies.append(y_pred[idx][:, classnum][:argmax])
    #     ax1.legend(frameon=True, fontsize=33)
    #     ax2.legend(frameon=True, fontsize=20.5, ncol=1, loc='right')
    #     ax2.set_xlabel("Days since trigger (rest frame)")  # , fontsize=18)
    #     ax1.set_ylabel("Relative Flux")  # , fontsize=15)
    #     ax2.set_ylabel("Class Probability")  # , fontsize=18)
    #     # ax1.set_ylim(-0.1, 1.1)
    #     # ax2.set_ylim(0, 1)
    #     mintime_lc = min([min(orig_lc_test[idx][pb]['time']) for pb in passbands])
    #     maxtime_lc = max([max(orig_lc_test[idx][pb]['time']) for pb in passbands])
    #     ax1.set_xlim(max(mintime_lc, -70), min(maxtime_lc, 80))
    #     # ax1.set_xlim(-70, 80)
    #     plt.setp(ax1.get_xticklabels(), visible=False)
    #     ax2.yaxis.set_major_locator(MaxNLocator(nbins=6, prune='upper'))  # added
    #     plt.tight_layout()
    #     fig.subplots_adjust(hspace=0)
    #     plt.savefig(os.path.join(fig_dir + '/lc_pred',
    #                              "classification_vs_time_{}_{}_{}_{}.pdf".format(idx, class_names[true_class], redshift,
    #                                                                              peakmjd - trigger_mjd)))
    #     plt.close()


# -*- coding: UTF-8 -*-
"""
Functions for the derivation of base features with LAobjects
"""

from __future__ import absolute_import
from __future__ import unicode_literals
import numpy as np
import astropy.table as at


class BaseMixin(object):
    """
    Methods to derive baseline features for LAobjects
    """

    def get_lc(self, recompute=False):
        """
        Most often, we want a lightcurve broken up passband by passband to
        compute features. This stores the input light curve in that format.
        Return a lightcurve suitable for computing features
        """
        outlc = getattr(self, '_outlc', None)
        if outlc is None or recompute:
            filters = self.filters
            outlc = {}

            # store the indices of each filter
            for i, pb in enumerate(self.filters):
                mask = np.where(self.passband == pb)[0]
                m2 = self.time[mask].argsort()
                ind = mask[m2]
                if len(ind) > 0:
                    outlc[pb] = ind
            self._outlc = outlc

        # for each of the default_cols, get the elements for t passband
        out = {}
        for pb, ind in outlc.items():
            out[pb] = [getattr(self, column)[ind] for column in self._default_cols]

        return out

    def get_lc_as_table(self):
        columns = ['passband', 'time', 'fluxUnred', 'fluxErrUnred', 'photflag']
        rename_columns = ['passband', 'time', 'flux', 'fluxErr', 'photflag']
        out = [getattr(self, col) for col in columns]

        out_table = at.Table(out, names=rename_columns)
        out_table = out_table.group_by('passband')

        return out_table





import os
import astrorapid.get_training_data
import astrorapid.get_custom_data
from astrorapid.custom_classifier import create_custom_classifier


def main():
    """ Train Neural Network classifier """

    script_dir = os.path.dirname(os.path.abspath(__file__))

    create_custom_classifier(get_data_func=astrorapid.get_training_data.get_real_ztf_training_data,
                             data_dir=os.path.join(script_dir, '..', 'data/real_ZTF_data_from_osc'),
                             class_nums=('Ia', 'Ia91T', 'Ia91bg', 'Iapec', 'Iacsm', 'Iax',
                                         'II', 'IIP', 'IIL', 'IIpec', 'IIn', 'IIb', 'Ib', 'Ibn', 'Ic', 'IcBL', 'Ibc',
                                         'CC', 'SLSN', 'SLSNI', 'SLSNII'),
                             class_name_map={'Ia': 'SNIa', 'Ia91T': 'SNIa', 'Ia91bg': 'SNIa', 'Iapec': 'SNIa',
                                             'Iacsm': 'SNIa', 'Iax': 'SNIa', 'II': 'CC', 'IIP': 'CC', 'IIL':
                                                 'CC', 'IIpec': 'CC', 'IIn': 'CC', 'IIb': 'CC',
                                             'Ib': 'CC', 'Ibn': 'CC', 'Ic': 'CC', 'IcBL': 'CC',
                                             'Ibc': 'CC', 'CC': 'CC',
                                             'SLSN': 'SLSN', 'SLSNI': 'SLSN', 'SLSNII': 'SLSN'},
                             reread_data=False,
                             train_size=0.6,
                             contextual_info=(),
                             passbands=('g', 'r'),
                             retrain_network=False,
                             train_epochs=300,
                             zcut=0.5,
                             bcut=False,
                             ignore_classes=('SLSN', 'SLSNI', 'SLSNII'),
                             nprocesses=1,
                             nchunks=10000,
                             otherchange='real-ztf-Ia-CC',
                             training_set_dir=os.path.join(script_dir, '..', 'training_set_files'),
                             save_dir=os.path.join(script_dir, '..', 'data/saved_real_ZTF_light_curves'),
                             fig_dir=os.path.join(script_dir, '..', 'training_set_files', 'Figures', 'ZTF_real_data-Ia-CC_no_redshift_epochs150'),
                             plot=True
                             )


if __name__ == '__main__':
    main()
